/*  core_misc.c - miscellaneous functions
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: core_misc.c,v 1.31 2003/03/19 06:17:31 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "shntool.h"
#include "wave.h"
#include "misc.h"
#include "binary.h"
#include "format.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

char *child_args[MAX_CHILD_ARGS];
int num_child_args = 0;

void my_snprintf(char *dest,int maxlen,char *formatstr, ...)
/* acts like snprintf, but makes 100% sure the string is NULL-terminated */
{
  va_list args;

  va_start(args,formatstr);

  my_vsnprintf(dest,maxlen,formatstr,args);

  dest[maxlen-1] = 0;

  va_end(args);
}

void length_to_str(wave_info *info)
/* converts length of file to a string in m:ss or m:ss.ff format */
{
  wlong newlength,rem1,rem2,frames;

  newlength = info->length;

  if (PROB_NOT_CD(info)) {
    if (0 != info->rate) {
      if ((info->data_size % info->rate) >= (info->rate / 2)) {
        newlength++;
      }
    }
    sprintf(info->m_ss,"%lu:%02lu",newlength/60,newlength%60);
  }
  else {
    rem1 = info->data_size % CD_RATE;
    rem2 = rem1 % CD_BLOCK_SIZE;

    frames = rem1 / CD_BLOCK_SIZE;
    if (rem2 >= (CD_BLOCK_SIZE / 2))
      frames++;

    if (frames == CD_BLOCKS_PER_SEC) {
      frames = 0;
      newlength++;
    }

    sprintf(info->m_ss,"%lu:%02lu.%02lu",newlength/60,newlength%60,frames);
  }
}

int filename_contains_a_dot(char *filename)
/* determines whether a filename contains a '.' in its base file name */
{
  char *slash,*dot;

  dot = strrchr(filename,'.');
  if (NULL == dot)
    return 0;

  slash = strrchr(filename,'/');
  if (NULL == slash)
    return 1;

  if (slash < dot)
    return 1;

  return 0;
}

#ifndef HAVE_SETLINEBUF
int setlinebuf(FILE *fp)
{
  (void) setvbuf(fp,
#ifdef SETVBUF_REVERSED
                 _IOLBF,NULL,
#else
                 NULL,_IOLBF,
#endif
                 0);
  return 0;
}
#endif

void close_and_wait(FILE *fd,int pid,int child_type)
{
  int gotpid,status;

  if (fd) {
    fclose(fd);
    fd = NULL;
  }

  if (NO_CHILD_PID != pid) {
    if (shntool_debug) {
      fprintf(stderr,"(waiting for %s process %d ... ",(CHILD_INPUT == child_type)?"input":"output",pid);
    }

    gotpid = (int)waitpid((pid_t)pid,&status,0);

    if (shntool_debug) {
      fprintf(stderr,"got pid %d",gotpid);

      fprintf(stderr," [%d",WIFEXITED(status));
      if (WIFEXITED(status))
        fprintf(stderr,"/%d",WEXITSTATUS(status));
      fprintf(stderr,"]");

      fprintf(stderr," [%d",WIFSIGNALED(status));
      if (WIFSIGNALED(status))
        fprintf(stderr,"/%d",WTERMSIG(status));
      fprintf(stderr,"]");

      fprintf(stderr," [%d",WIFSTOPPED(status));
      if (WIFSTOPPED(status))
        fprintf(stderr,"/%d",WSTOPSIG(status));
      fprintf(stderr,"]");

      fprintf(stderr,")\n");
    }

    if ((CHILD_OUTPUT == child_type) && WIFEXITED(status) && WEXITSTATUS(status)) {
      warning("child process %d had non-zero exit status %d -\n"
              "output file may not have been written correctly",
              pid,WEXITSTATUS(status));
    }
  }
}

int spawn(FILE **readpipe,FILE **writepipe,int child_type)
/* forks off a process running the command cmd, and sets up read/write
 * pipes for two-way communication with that process
 */
{
  int childpid, pipe1[2], pipe2[2];
  int i,nullfd,max_fds;
  char **real_args;

  if (NULL == (real_args = malloc(num_child_args * (sizeof(char *) + 1)))) {
    perror("shntool: malloc");
    error("error while allocating space for child process arguments");
  }

  for (i=0;i<num_child_args;i++)
    real_args[i] = child_args[i];
  real_args[num_child_args] = NULL;

  for (i=0;i<MAX_CHILD_ARGS;i++)
    child_args[i] = NULL;
  num_child_args = 0;

  if ((pipe(pipe1) < 0) || (pipe(pipe2) < 0)) {
    perror("shntool: pipe");
    error("error while creating pipes for two-way communication with child process, see above");
   }

  switch ((childpid = fork())) {
    case -1:
      /* fork failed */
      perror("shntool: fork");
      error("error while forking child process, see above");
      break;
    case 0:
      /* child */

      close(pipe1[1]);
      close(pipe2[0]);
      /* Read from parent on pipe1[0], write to parent on pipe2[1]. */
      dup2(pipe1[0],0);
      dup2(pipe2[1],1);
      close(pipe1[0]);
      close(pipe2[1]);

      /* make sure stderr is connected to /dev/null, or closed if that fails */
      close(2);
      nullfd = open("/dev/null",O_WRONLY);
      if (nullfd > 0 && 2 != nullfd)
        dup2(nullfd,2);

#ifdef __CYGWIN__
      setmode(0,_O_BINARY);
      setmode(1,_O_BINARY);
      setmode(2,_O_BINARY);
#endif

#ifdef HAVE_SYSCONF
      max_fds = sysconf(_SC_OPEN_MAX);
#else
      max_fds = 1024;
#endif

      /* close all other file descriptors */
      for (i=3;i<max_fds;i++)
        close(i);

      if (execvp(real_args[0],real_args) < 0) {
        perror("shntool: execvp");
        error("error while executing '%s', see above",real_args[0]);
      }
      break;
    default:
      /* parent */
      close(pipe1[0]);
      close(pipe2[1]);
      /* Write to child on pipe1[1], read from child on pipe2[0]. */
      *readpipe = fdopen(pipe2[0], "r");
      *writepipe = fdopen(pipe1[1], "w");
      setlinebuf(*writepipe);
      SETBINARY_IN(*readpipe);
      SETBINARY_OUT(*writepipe);
      if (shntool_debug) {
        fprintf(stderr,"(spawned %s process '%s' with pid %d and arguments:",(CHILD_INPUT == child_type)?"input":"output",real_args[0],childpid);
        for (i=1;real_args[i];i++)
          fprintf(stderr," \"%s\"",real_args[i]);
        fprintf(stderr,")\n");
      }
      free(real_args);
      break;
    }

  return childpid;
}

format_module *output_format_init(int argc,char **argv,int *argpos)
{
  int i;
  format_module *op = NULL;

  (*argpos)++;

  if (argc <= *argpos)
    help("missing output file format");

  for (i=0;formats[i];i++) {
    if (0 == strcmp(argv[*argpos],formats[i]->name)) {
      if (NULL == formats[i]->open_output)
        help("file format '%s' does not support output",formats[i]->name);
      op = formats[i];
      break;
    }
  }

  if (NULL == op)
    help("invalid output file format: %s",argv[*argpos]);

  if (op->get_arguments)
    op->get_arguments(argc,argv,argpos);

  return op;
}

void arg_add(char *arg)
{
  if (num_child_args >= MAX_CHILD_ARGS)
    error("too many arguments specified - limit is %d",MAX_CHILD_ARGS);

  child_args[num_child_args] = arg;
  num_child_args++;
}

void alter_file_order(wave_info **filenames,int numfiles)
/* a simple menu system to change the order of the files as given on the command line */
{
  int i,current,a,b;
  char response[BUF_SIZE],
       *p,
       *args[3];
  wave_info *tmp;

  printf("\nEntering file order editor.\n");

  while (1) {
    printf("\n");
    printf("File list:\n");
    printf("\n");
    for (i=0;i<numfiles;i++)
      printf("%2d: %s\n",i+1,filenames[i]->filename);
    printf("\n");
    printf("Commands:\n");
    printf("\n");
    printf("  swap a b   (swaps positions a and b)\n");
    printf("  move a b   (moves position a to position b, shifting everything in between)\n");
    printf("  begin a    (moves position a to the beginning of the list)\n");
    printf("  end a      (moves position a to the end of the list)\n");
    printf("  done       (quits this editor, and continues with processing)\n");
    printf("  quit       (quits %s [you can also use Ctrl-C])\n",progname);
    printf("\n? ");
    fgets(response,BUF_SIZE-1,stdin);
    response[strlen(response)-1] = '\0';
    p = response;
    current = 0;
    args[0] = NULL;
    args[1] = NULL;
    args[2] = NULL;
    while ('\0' != *p) {
      while (' ' == *p || '\t' == *p || '\n' == *p) {
        *p = '\0';
        p++;
      }
      if ('\0' == *p)
        break;
      if (3 > current)
        args[current++] = p;
      while ('\0' != *p && ' ' != *p && '\t' != *p && '\n' != *p)
        p++;
    }
    if (NULL == args[0])
      args[0] = "";
    if (0 == strcmp(args[0],"quit")) {
      printf("\nQuitting %s.\n",progname);
      exit(1);
    }
    else if (0 == strcmp(args[0],"done")) {
      printf("\nExiting file order editor.\n");
      break;
    }
    else if (0 == strcmp(args[0],"swap")) {
      if (NULL == args[2])
        printf("\nNot enough positions supplied to the 'swap' command\n");
      else {
        a = atoi(args[1]);
        b = atoi(args[2]);
        if ((a<1 || a>numfiles) || (b<1 || b>numfiles))
          printf("\nOne or more of your entries are out of range.  Try again.\n");
        else {
          printf("\nSwapping positions %d and %d...\n",a,b);
          tmp = filenames[a-1];
          filenames[a-1] = filenames[b-1];
          filenames[b-1] = tmp;
        }
      }
    }
    else if (0 == strcmp(args[0],"move")) {
      if (NULL == args[2])
        printf("\nNot enough positions supplied to the 'move' command\n");
      else {
        a = atoi(args[1]);
        b = atoi(args[2]);
        if ((a<1 || a>numfiles) || (b<1 || b>numfiles))
          printf("\nOne or more of your entries are out of range.  Try again.\n");
        else {
          printf("\nMoving position %d to position %d...\n",a,b);
          tmp = filenames[a-1];
          if (a < b) {
            for (i=a-1;i<b-1;i++)
              filenames[i] = filenames[i+1];
          }
          else if (a > b) {
            for (i=a-1;i>b-1;i--)
              filenames[i] = filenames[i-1];
          }
          filenames[b-1] = tmp;
        }
      }
    }
    else if (0 == strcmp(args[0],"begin")) {
      if (NULL == args[1])
        printf("\nMissing a position for the 'begin' command.\n");
      else {
        a = atoi(args[1]);
        if ((a<1 || a>numfiles))
          printf("\nYour entry is out of range.  Try again.\n");
        else {
          printf("\nMoving position %d to the beginning of the list\n",a);
          tmp = filenames[a-1];
          for (i=a-1;i>0;i--)
            filenames[i] = filenames[i-1];
          filenames[0] = tmp;
        }
      }
    }
    else if (0 == strcmp(args[0],"end")) {
      if (NULL == args[1])
        printf("\nMissing a position for the 'end' command.\n");
      else {
        a = atoi(args[1]);
        if ((a<1 || a>numfiles))
          printf("\nYour entry is out of range.  Try again.\n");
        else {
          printf("\nMoving position %d to the end of the list...\n",a);
          tmp = filenames[a-1];
          for (i=a-1;i<numfiles-1;i++)
            filenames[i] = filenames[i+1];
          filenames[numfiles-1] = tmp;
        }
      }
    }
    else {
      printf("\nUnrecognized command: '%s'.\n",args[0]);
    }
  }

  printf("\n");
}

void internal_version()
{
  printf("%s%s %s\n",fullprogname,(NULL == progmode)?"":" mode module",version);
  printf("%s %s\n\n",copyright,author);
  printf("shorten utilities pages:\n\n%s\n",urls);

  printf(
         "\n"
         "This program comes with ABSOLUTELY NO WARRANTY.\n"
         "This is free software, and you are welcome to redistribute it\n"
         "under certain conditions.  See the file COPYING for details.\n"
         "\n"
        );

  exit(0);
}
