/*  core_wave.c - functions for parsing WAVE headers
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: core_wave.c,v 1.33 2003/02/22 09:40:35 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include "shntool.h"
#include "convert.h"
#include "fileio.h"
#include "misc.h"
#include "wave.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef HAVE_STRERROR
extern char *sys_errlist[];
#endif

int is_valid_file(wave_info *info)
/* determines whether the given filename (info->filename) is a regular file, and is readable */
{
  struct stat sz;
  FILE *f;

  if (stat(info->filename,&sz)) {
    if (errno == ENOENT)
      warning("cannot open '%s' because it does not exist",info->filename);
    else if (errno == EACCES)
      warning("cannot open '%s' due to insufficient permissions",info->filename);
    else if (errno == EFAULT)
      warning("cannot open '%s' due to bad address",info->filename);
    else if (errno == ENOMEM)
      warning("cannot open '%s' because the kernel ran out of memory",info->filename);
    else if (errno == ENAMETOOLONG)
      warning("cannot open '%s' because the file name is too long",info->filename);
    else
      warning("cannot open '%s' due to an unknown problem",info->filename);
    return 0;
  }
  if (0 == S_ISREG(sz.st_mode)) {
    if (S_ISLNK(sz.st_mode))
      warning("'%s' is a symbolic link, not a regular file",info->filename);
    else if (S_ISDIR(sz.st_mode))
      warning("'%s' is a directory, not a regular file",info->filename);
    else if (S_ISCHR(sz.st_mode))
      warning("'%s' is a character device, not a regular file",info->filename);
    else if (S_ISBLK(sz.st_mode))
      warning("'%s' is a block device, not a regular file",info->filename);
    else if (S_ISFIFO(sz.st_mode))
      warning("'%s' is a fifo, not a regular file",info->filename);
    else if (S_ISSOCK(sz.st_mode))
      warning("'%s' is a socket, not a regular file",info->filename);
    return 0;
  }
  info->actual_size = (wlong)sz.st_size;

  if (NULL == (f = fopen(info->filename,"rb"))) {
    warning("could not open '%s': %s",info->filename,
#ifdef HAVE_STRERROR
          strerror(errno)
#else
          sys_errlist[errno]
#endif
          );
    return 0;
  }
  fclose(f);

  return 1;
}

int aiff_input_kluge(unsigned char *header,wave_info *info)
/* kluge to set data_size and chunk_size, since sox can't fseek() on
 * stdout to backfill these values.  This relies on the value of the
 * data size in the SSND chunk.
 */
{
  FILE *f;
  unsigned long be_long=0;
  unsigned char tag[4];

  if (NULL == (f = fopen(info->filename,"rb")))
    return 0;

  /* discard FORM, size and form type */
  if (0 == read_be_long(f,&be_long) || 0 == read_be_long(f,&be_long) || 0 == read_be_long(f,&be_long)) {
    fclose(f);
    return 0;
  }

  while (1) {
    /* read chunk id */
    if (0 == read_tag(f,tag)) {
      fclose(f);
      return 0;
    }

    if (0 == tagcmp(tag,AIFF_SSND))
      break;

    /* not SSND, so read size of this chunk and skip it */
    if (0 == read_be_long(f,&be_long)) {
      fclose(f);
      return 0;
    }

    fseek(f,(long)be_long,SEEK_CUR);
  }

  /* now read SSND size */
  if (0 == read_be_long(f,&be_long)) {
    fclose(f);
    return 0;
  }

  fclose(f);

  /* subtract 8 bytes, because SSND size includes offset and blocksize (4 bytes each) */
  be_long -= 8;

  /* set proper data size */
  info->data_size = be_long;

  /* now set chunk size based on data size, and the canonical WAVE header size, which sox generates */
  info->chunk_size = info->data_size + CANONICAL_HEADER_SIZE - 8;

  make_canonical_header(header,info);

  return 1;
}

int ape_input_kluge(unsigned char *header,wave_info *info)
{
  if (info->chunk_size < info->data_size)
    info->chunk_size = info->header_size + info->data_size - 8;

  put_data_size(header,info->header_size,info->data_size);

  return 1;
}

int do_header_kluges(unsigned char *header,wave_info *info)
{
  int failed = 0;

  if (0 == strcmp(info->input_format->name,"aiff") && 0 == aiff_input_kluge(header,info))
    failed = 1;
  else if (0 == strcmp(info->input_format->name,"ape") && 0 == ape_input_kluge(header,info))
    failed = 1;

  if (failed) {
    warning("while processing '%s': %s to wav input kluge failed",info->filename,info->input_format->name);
    return 0;
  }

  return 1;
}

int verify_wav_header_internal(wave_info *info,int verbose)
/* verifies that data coming in on the file descriptor info->input describes a valid WAVE header */
{
  unsigned long le_long=0,bytes;
  unsigned char buf[2];
  unsigned char tag[4];
  int header_len = 0;

  /* look for "RIFF" in header */
  if (0 == read_tag(info->input,tag) || tagcmp(tag,WAVE_RIFF)) {
    if (0 == tagcmp(tag,AIFF_FORM)) {
      if (verbose) warning("while processing '%s': file contains AIFF data, which is currently not supported",info->filename);
    }
    else {
      if (verbose) warning("while processing '%s': WAVE header is missing RIFF tag - possible truncated/corrupt file",info->filename);
    }
    return 0;
  }

  if (0 == read_le_long(info->input,&info->chunk_size)) {
    if (verbose) warning("while processing '%s': could not read chunk size from WAVE header",info->filename);
    return 0;
  }

  /* look for "WAVE" in header */
  if (0 == read_tag(info->input,tag) || tagcmp(tag,WAVE_WAVE)) {
    if (verbose) warning("while processing '%s': WAVE header is missing WAVE tag",info->filename);
    return 0;
  }

  header_len += 12;

  for (;;) {
    if (0 == read_tag(info->input,tag) || 0 == read_le_long(info->input,&le_long)) {
      if (verbose) warning("while processing '%s': reached end of file while looking for fmt tag",info->filename);
      return 0;
    }
    header_len += 8;
    if (0 == tagcmp(tag,WAVE_FMT))
      break;
    bytes = le_long;
    while (bytes > 0) {
      if (fread(buf, 1, 1, info->input) != 1) {
        if (verbose) warning("while processing '%s': reached end of file while jumping ahead %lu bytes during search for fmt tag",info->filename,le_long);
        return 0;
      }
      bytes--;
    }
    header_len += le_long;
  }

  if (le_long < 16) {
    if (verbose) warning("while processing '%s': fmt chunk in WAVE header was too short",info->filename);
    return 0;
  }

  /* now we read the juicy stuff */
  if (0 == read_le_short(info->input,&info->wave_format)) {
    if (verbose) warning("while processing '%s': reached end of file while reading format",info->filename);
    return 0;
  }

  switch (info->wave_format) {
    case WAVE_FORMAT_PCM:
      break;
    default:
      if (verbose) warning("while processing '%s': unsupported format 0x%04x (%s) - only PCM data is supported at this time",
            info->filename,info->wave_format,format_to_str(info->wave_format));
      return 0;
  }

  if (0 == read_le_short(info->input,&info->channels)) {
    if (verbose) warning("while processing '%s': reached end of file while reading channels",info->filename);
    return 0;
  }

  if (0 == read_le_long(info->input,&info->samples_per_sec)) {
    if (verbose) warning("while processing '%s': reached end of file while reading samples/sec",info->filename);
    return 0;
  }

  if (0 == read_le_long(info->input,&info->avg_bytes_per_sec)) {
    if (verbose) warning("while processing '%s': reached end of file while reading average bytes/sec",info->filename);
    return 0;
  }

  if (0 == read_le_short(info->input,&info->block_align)) {
    if (verbose) warning("while processing '%s': reached end of file while reading block align",info->filename);
    return 0;
  }

  if (0 == read_le_short(info->input,&info->bits_per_sample)) {
    if (verbose) warning("while processing '%s': reached end of file while reading bits/sample",info->filename);
    return 0;
  }

  header_len += 16;

  le_long -= 16;

  if (0 != le_long) {
    bytes = le_long;
    while (bytes > 0) {
      if (fread(buf, 1, 1, info->input) != 1) {
        if (verbose) warning("while processing '%s': reached end of file while jumping ahead %lu bytes",info->filename,le_long);
        return 0;
      }
      bytes--;
    }
    header_len += le_long;
  }

  /* now let's look for the data chunk.  Following the string "data" is the
     length of the following WAVE data. */
  for (;;) {
    if (0 == read_tag(info->input,tag) || 0 == read_le_long(info->input,&le_long)) {
      if (verbose) warning("while processing '%s': reached end of file while looking for data tag",info->filename);
      return 0;
    }
    header_len += 8;
    if (0 == tagcmp(tag,WAVE_DATA))
      break;
    bytes = le_long;
    while (bytes > 0) {
      if (fread(buf, 1, 1, info->input) != 1) {
        if (verbose) warning("while processing '%s': end of file encountered when jumping ahead %lu bytes while looking for data tag",info->filename,le_long);
        return 0;
      }
      bytes--;
    }
    header_len += le_long;
  }

  info->data_size = le_long;

  info->header_size = header_len;

  if (0 == do_header_kluges(NULL,info))
    return 0;

  info->rate = ((wint)info->samples_per_sec * (wint)info->channels * (wint)info->bits_per_sample) / 8;

  info->total_size = info->chunk_size + 8;

  info->length = info->data_size / info->rate;

  if (info->channels == CD_CHANNELS &&
      info->bits_per_sample == CD_BITS_PER_SAMPLE &&
      info->samples_per_sec == CD_SAMPLES_PER_SEC &&
/*
  removed, since this doesn't seem to matter
      info->block_align == CD_BLOCK_ALIGN &&
*/
      info->avg_bytes_per_sec == CD_RATE &&
      info->rate == CD_RATE)
  {
    if (info->data_size < CD_MIN_BURNABLE_SIZE)
      info->problems |= PROBLEM_CD_BUT_TOO_SHORT;
    if (info->data_size % CD_BLOCK_SIZE != 0)
      info->problems |= PROBLEM_CD_BUT_BAD_BOUND;
  }
  else
    info->problems |= PROBLEM_NOT_CD_QUALITY;

  if (info->header_size != CANONICAL_HEADER_SIZE)
    info->problems |= PROBLEM_HEADER_NOT_CANONICAL;

  if ((wlong)info->header_size + info->data_size > info->total_size)
    info->problems |= PROBLEM_HEADER_INCONSISTENT;

  if ((0 == info->input_format->is_compressed) && (0 == info->input_format->is_translated) && (info->total_size < info->actual_size))
    info->problems |= PROBLEM_JUNK_APPENDED;

  if ((0 == info->input_format->is_compressed) && (0 == info->input_format->is_translated) && (info->total_size > info->actual_size))
    info->problems |= PROBLEM_MAY_BE_TRUNCATED;

  if ((wlong)info->header_size + info->data_size < info->total_size)
    info->problems |= PROBLEM_EXTRA_CHUNKS;

  length_to_str(info);

  /* header looks ok */
  return 1;
}

wave_info *new_wave_info(char *filename)
/* if filename is NULL, return a fresh wave_info struct with all data zero'd out.
 * Otherwise, check that the file referenced by filename exists, is readable, and
 * contains a valid WAVE header - if so, return a wave_info struct filled out with
 * the values in the WAVE header, otherwise return NULL.
 */
{
  int i, kidpid;
  wave_info *info;
  char buf[4];

  if (NULL == (info = malloc(sizeof(wave_info)))) {
    warning("could not allocate memory for WAVE info struct");
    goto invalid_wave_data;
  }

  /* set defaults */
  memset((void *)info,0,sizeof(wave_info));

  if (NULL == filename)
    return info;

  info->filename = filename;

  if (0 == is_valid_file(info))
    goto invalid_wave_data;

  for (i=0;formats[i];i++)
    if (formats[i]->open_input && formats[i]->is_our_file(filename)) {

      info->input_format = formats[i];

      if (NULL == (info->input = formats[i]->open_input(filename,&kidpid))) {
        warning("could not open '%s', though the '%s' format module claims to support it",filename,formats[i]->name);
        goto invalid_wave_data;
      }

      if (1 != fread(&buf,1,1,info->input)) {
        if (info->input_format->decoder)
          warning("could not read data from file '%s' after opening it with the '%s' format\n"
                  "module - possible truncated/corrupt file, or helper program '%s' could not be found",
                  filename,info->input_format->name,info->input_format->decoder);
        else
          warning("could not read 1 byte from file '%s' after opening it with the '%s' format\n"
                  "module - possible truncated/corrupt file",filename,formats[i]->name);
        goto invalid_wave_data;
      }
      close_and_wait_input(info->input,kidpid);

      if (NULL == (info->input = formats[i]->open_input(filename,&kidpid))) {
        warning("could not re-open '%s' for input while attempting to verify WAVE header");
        goto invalid_wave_data;
      }

      if (0 == verify_wav_header_internal(info,1))
        goto invalid_wave_data;

      close_and_wait_input(info->input,kidpid);
      info->input = NULL;

      /* success */
      return info;
    }

  /* if we got here, no file format modules claimed to handle the file */

  warning("file '%s' is not handled by any of the builtin format modules",filename);

invalid_wave_data:

  if (info) {
    if (info->input) {
      close_and_wait_input(info->input,kidpid);
      info->input = NULL;
    }
    free(info);
  }
  return NULL;
}

void make_canonical_header(unsigned char *header,wave_info *info)
/* constructs a canonical WAVE header from the values in the wave_info struct */
{
  if (NULL == header)
    return;

  tagcpy(header,WAVE_RIFF);
  ulong_to_uchar_le(header+4,info->chunk_size);
  tagcpy(header+8,WAVE_WAVE);
  tagcpy(header+12,WAVE_FMT);
  ulong_to_uchar_le(header+16,0x00000010);
  ushort_to_uchar_le(header+20,info->wave_format);
  ushort_to_uchar_le(header+22,info->channels);
  ulong_to_uchar_le(header+24,info->samples_per_sec);
  ulong_to_uchar_le(header+28,info->avg_bytes_per_sec);
  ushort_to_uchar_le(header+32,info->block_align);
  ushort_to_uchar_le(header+34,info->bits_per_sample);
  tagcpy(header+36,WAVE_DATA);
  ulong_to_uchar_le(header+40,info->data_size);
}

char *format_to_str(wshort format)
{
  switch (format) {
    case WAVE_FORMAT_UNKNOWN:
      return "Microsoft Official Unknown";
    case WAVE_FORMAT_PCM:
      return "Microsoft PCM";
    case WAVE_FORMAT_ADPCM:
      return "Microsoft ADPCM";
    case WAVE_FORMAT_IEEE_FLOAT:
      return "IEEE Float";
    case WAVE_FORMAT_ALAW:
      return "Microsoft A-law";
    case WAVE_FORMAT_MULAW:
      return "Microsoft U-law";
    case WAVE_FORMAT_OKI_ADPCM:
      return "OKI ADPCM format";
    case WAVE_FORMAT_IMA_ADPCM:
      return "IMA ADPCM";
    case WAVE_FORMAT_DIGISTD:
      return "Digistd format";
    case WAVE_FORMAT_DIGIFIX:
      return "Digifix format";
    case WAVE_FORMAT_DOLBY_AC2:
      return "Dolby AC2";
    case WAVE_FORMAT_GSM610:
      return "GSM 6.10";
    case WAVE_FORMAT_ROCKWELL_ADPCM:
      return "Rockwell ADPCM";
    case WAVE_FORMAT_ROCKWELL_DIGITALK:
      return "Rockwell DIGITALK";
    case WAVE_FORMAT_G721_ADPCM:
      return "G.721 ADPCM";
    case WAVE_FORMAT_G728_CELP:
      return "G.728 CELP";
    case WAVE_FORMAT_MPEG:
      return "MPEG";
    case WAVE_FORMAT_MPEGLAYER3:
      return "MPEG Layer 3";
    case WAVE_FORMAT_G726_ADPCM:
      return "G.726 ADPCM";
    case WAVE_FORMAT_G722_ADPCM:
      return "G.722 ADPCM";
  }
  return "Unknown";
}

void put_data_size(unsigned char *header,int header_size,unsigned long new_size)
/* replaces the size reported in the "data" chunk of the wave header with the new size -
   also updates chunk size at beginning of the wave header */
{
  if (NULL == header)
    return;

  ulong_to_uchar_le(header+header_size-4,new_size);
  ulong_to_uchar_le(header+4,new_size+header_size-8);
}

void pad_file(FILE *out,int bytes)
/* writes the specified number of zero bytes to the file descriptor given */
{
  int i;

  for (i=0;i<bytes;i++)
    fprintf(out,"%c",(unsigned char)0);
}
