/*  mode_cat.c - cat mode module
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: mode_cat.c,v 1.16 2003/03/19 05:49:02 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

static int cat_header = 1;
static int cat_data = 1;
static int cat_extra = 1;

static void cat_main(int,char **,int);

mode_module mode_cat = {
  "cat",
  "shncat",
  "writes PCM WAVE data from one or more files to stdout",
  cat_main
};

static void show_usage()
{
  printf("Usage: %s [OPTIONS] [file]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -nh  specifies that the WAVE header should be suppressed from the output\n");
  printf("\n");
  printf("  -nd  specifies that the WAVE data should be suppressed from the output\n");
  printf("\n");
  printf("  -nr  specifies that extra RIFF chunks should be suppressed from the output\n");
  printf("\n");
  printf("  --   indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v   shows version information\n");
  printf("  -h   shows this help screen\n");
  printf("\n");
  printf("  If no file name is given, then a file name is read from standard input.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -nh filename.wav\n",fullprogname);
  printf("\n");
  printf("  %% echo filename.shn | %s\n",fullprogname);
  printf("\n");

  exit(0);
}

static void parse(int argc,char **argv,int begin,int *first_arg)
{
  int start;

  start = begin;

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage();
    else if (argc > start && 0 == strcmp(argv[start],"-nh"))
      cat_header = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-nd"))
      cat_data = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-nr"))
      cat_extra = 0;
    else
      help("unknown argument: %s",argv[start]);
    start++;
  }

  if (0 == cat_header && 0 == cat_data && 0 == cat_extra)
    help("nothing to do if -nh, -nd and -nr are specified");

  *first_arg = start;
}

static void cat_file(wave_info *info)
{
  unsigned char *header;
  wlong extra_bytes;
  FILE *f;
  int kidpid;

  extra_bytes = info->total_size - (info->data_size + info->header_size);

  if (0 == cat_header && 0 == cat_data && 0 == extra_bytes)
    error("file contains no extra RIFF chunks - nothing to do");

  if (NULL == (info->input = info->input_format->open_input(info->filename,&kidpid)))
    error("could not reopen file '%s' for input",info->filename);

  if (NULL == (header = malloc(info->header_size * sizeof(unsigned char))))
    error("could not allocate %d bytes for WAVE header",info->header_size);

  if (read_n_bytes(info->input,header,info->header_size) != info->header_size)
    error("error while reading in %d-byte header",info->header_size);

  if (1 == cat_header) {
    if (0 == do_header_kluges(header,info))
      error("could not fix WAVE header, aborting.");

    if (write_n_bytes(stdout,header,info->header_size) != info->header_size)
      error("error while writing %d bytes of header",info->header_size);
  }

  if (1 == cat_data) {
    if (transfer_n_bytes(info->input,stdout,info->data_size) != info->data_size)
      error("error while transferring %lu bytes of data",info->data_size);
  }

  if (1 == cat_extra) {
    if (NULL == (f = fopen("/dev/null","wb")))
      error("could not open /dev/null for output");

    if (0 == cat_data) {
      if (transfer_n_bytes(info->input,f,info->data_size) != info->data_size)
        error("error while transferring %lu bytes of data",info->data_size);
    }

    if (transfer_n_bytes(info->input,stdout,extra_bytes) != extra_bytes)
      error("error while transferring %lu bytes of data",extra_bytes);
  }

  free(header);

  close_and_wait_input(info->input,kidpid);
}

static void process_file(char *filename)
{
  wave_info *info;

  if (NULL == (info = new_wave_info(filename)))
    return;

  cat_file(info);

  free(info);
}

static void process(int argc,char **argv,int start)
{
  char filename[FILENAME_SIZE];
  int i;

  if (argc < start + 1) {
    /* no filenames were given, so we're reading files from standard input.  */
    fgets(filename,FILENAME_SIZE-1,stdin);
    while (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file(filename);
      fgets(filename,FILENAME_SIZE-1,stdin);
    }
  }
  else
    for (i=start;i<argc;i++)
      process_file(argv[i]);
}

static void cat_main(int argc,char **argv,int start)
{
  int first_arg;

  parse(argc,argv,start,&first_arg);

  process(argc,argv,first_arg);
}
