/*  mode_conv.c - conv mode module
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: mode_conv.c,v 1.23 2003/02/17 04:10:42 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

static int output_dir_flag = 0;

static char *output_directory = ".";

static format_module *op = NULL;

static void conv_main(int,char **,int);

mode_module mode_conv = {
  "conv",
  "shnconv",
  "converts files from one format to another",
  conv_main
};

static void show_usage()
{
  int i,j=0;

  printf("Usage: %s [OPTIONS] [files]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt  specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (formats[i]->open_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir  specifies alternate output directory (default is same as input file)\n");
  printf("\n");
  printf("  --      indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v      shows version information\n");
  printf("  -h      shows this help screen\n");
  printf("\n");
  printf("  If no file names are given, then file names are read from standard input.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -o shn filename.wav\n",fullprogname);
  printf("\n");
  printf("  %% find / -name \\*.shn | %s -o flac\n",fullprogname);
  printf("\n");

  exit(0);
}

static void parse(int argc,char **argv,int begin,int *first_arg)
{
  int start,i,found_output = -1;

  start = begin;

  for (i=0;formats[i];i++)
    if (formats[i]->open_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output)
    error("no output formats found, cannot continue");

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage();
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == output_dir_flag)
        help("too many output directories specified");
      start++;
      output_dir_flag = 1;
      if (argc <= start)
        help("missing output directory");
      output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != op)
        help("too many output file formats specified");
      op = output_format_init(argc,argv,&start);
    }
    else {
      help("unknown argument: %s",argv[start]);
    }
    start++;
  }

  if (NULL == op) {
    op = formats[found_output];
    warning("no output format specified - defaulting to %s",op->name);
  }

  *first_arg = start;
}

static void make_outfilename(wave_info *info,char *outfilename)
{
  char *ext,*basename,infilename[FILENAME_SIZE];

  strcpy(infilename,info->filename);

  if ((basename = strrchr(infilename,'/')))
    basename++;
  else
    basename = infilename;

  if ((ext = strrchr(basename,'.')) && (0 == strcmp(ext+1,info->input_format->extension)))
    *ext = '\0';

  if (1 == output_dir_flag)
    my_snprintf(outfilename,FILENAME_SIZE,"%s/%s.%s",output_directory,basename,op->extension);
  else
    my_snprintf(outfilename,FILENAME_SIZE,"%s.%s",infilename,op->extension);
}

static void conv_file(wave_info *info)
{
  int kidpid_in,kidpid_out;
  wlong extra_bytes;
  FILE *out;
  char outfilename[FILENAME_SIZE];
  unsigned char *header;

  extra_bytes = info->total_size - (info->data_size + info->header_size);

  if (NULL == (info->input = info->input_format->open_input(info->filename,&kidpid_in)))
    error("could not open file '%s' for input",info->filename);

  make_outfilename(info,outfilename);

  if (NULL == (out = op->open_output(outfilename,&kidpid_out)))
    error("could not open file '%s' for put",outfilename);

  printf("converting '%s' to '%s' ... ",info->filename,outfilename);
  fflush(stdout);

  if (NULL == (header = malloc(info->header_size * sizeof(unsigned char))))
    error("could not allocate %d bytes for WAVE header",info->header_size);

  if (read_n_bytes(info->input,header,info->header_size) != info->header_size)
    error("error while discarding %d-byte header",info->header_size);

  if (0 == do_header_kluges(header,info))
    error("could not fix WAVE header, aborting.");

  if ((info->header_size > 0) && write_n_bytes(out,header,info->header_size) != info->header_size)
    error("error while writing %d-byte header",info->header_size);

  if ((info->data_size > 0) && (transfer_n_bytes(info->input,out,info->data_size) != info->data_size))
    error("error while transferring %lu-byte data chunk",info->data_size);

  if ((extra_bytes > 0) && (transfer_n_bytes(info->input,out,extra_bytes) != extra_bytes))
    error("error while transferring %lu extra bytes",extra_bytes);

  close_and_wait_input(info->input,kidpid_in);
  close_and_wait_output(out,kidpid_out);

  printf("done.\n");
}

static void process_file(char *filename)
{
  wave_info *info;

  if (NULL == (info = new_wave_info(filename)))
    return;

  conv_file(info);

  free(info);
}

static void process(int argc,char **argv,int start)
{
  char filename[FILENAME_SIZE];
  int i;

  if (argc < start + 1) {
    /* no filename was given, so we're reading one file name from standard input. */
    fgets(filename,FILENAME_SIZE-1,stdin);
    while (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file(filename);
      fgets(filename,FILENAME_SIZE-1,stdin);
    }
  }
  else {
    for (i=start;i<argc;i++)
      process_file(argv[i]);
  }
}

static void conv_main(int argc,char **argv,int start)
{
  int first_arg;

  parse(argc,argv,start,&first_arg);

  process(argc,argv,first_arg);
}
