/*  mode_split.c - split mode module
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: mode_split.c,v 1.20 2003/03/17 01:56:46 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define SPLIT_PREFIX "split-track"
#define MAX_PIECES   256

static int split_point_file_flag = 0;
static int prefix_flag = 0;
static int output_dir_flag = 0;
static int numfiles = 0;
static int preview = 0;
static int offset_flag = 0;
static int offset = 1;
static int input_is_cd_quality = 0;

static char *split_point_file = NULL;
static char *prefix = SPLIT_PREFIX;
static char *output_directory = ".";

static wave_info *files[MAX_PIECES];

static format_module *op = NULL;

static void split_main(int,char **,int);

mode_module mode_split = {
  "split",
  "shnsplit",
  "splits PCM WAVE data from one file into smaller files",
  split_main
};

static void show_usage()
{
  int i, j = 0;

  printf("Usage: %s [OPTIONS] file\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt   specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (formats[i]->open_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir   specifies alternate output directory (default is current directory)\n");
  printf("\n");
  printf("  -f file  specifies a file from which to read split point data\n");
  printf("\n");
  printf("  -n name  specifies a prefix for output files (default is \"%s\")\n",SPLIT_PREFIX);
  printf("\n");
  printf("  -c num   start counting from num when naming output files (default is 1)\n");
  printf("\n");
  printf("  -p       preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --       indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v       shows version information\n");
  printf("  -h       shows this help screen\n");
  printf("\n");
  printf("  If no split point file is given, then split points are read from standard\n");
  printf("  input.  See the man page (or README.txt) for more details about split points.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -f offsets.txt -n bandYYYY-MM-DDd2t bigfile.wav\n",fullprogname);
  printf("\n");
  printf("  %% %s -p -d disc1 bigfile.shn < splitpoints\n",fullprogname);
  printf("\n");

  exit(0);
}

static void parse(int argc,char **argv,int begin,int *first_arg)
{
  int start,i,found_output = -1;

  start = begin;

  for (i=0;formats[i];i++)
    if (formats[i]->open_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output)
    error("no output formats found, cannot continue");

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage();
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-c")) {
      if (1 == offset_flag)
        help("too many starting counts specified");
      start++;
      offset_flag = 1;
      if (argc <= start)
        help("missing starting count");
      offset = atoi(argv[start]);
      if (offset < 0)
        help("starting count cannot be negative");
    }
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == output_dir_flag)
        help("too many output directories specified");
      start++;
      output_dir_flag = 1;
      if (argc <= start)
        help("missing output directory");
      output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-n")) {
      if (1 == prefix_flag)
        help("too many prefixes files specified");
      start++;
      prefix_flag = 1;
      if (argc <= start)
        help("missing prefix");
      prefix = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-f")) {
      if (1 == split_point_file_flag)
        help("too many split point files specified");
      start++;
      split_point_file_flag = 1;
      if (argc <= start)
        help("missing split point file");
      split_point_file = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != op)
        help("too many output file formats specified");
      op = output_format_init(argc,argv,&start);
    }
    else {
      help("unknown argument: %s",argv[start]);
    }
    start++;
  }

  if (start != argc - 1)
    help("need exactly one file to process");

  if (NULL == op) {
    op = formats[found_output];
    warning("no output format specified - defaulting to %s",op->name);
  }

  *first_arg = start;
}

static void make_outfilename(char *outfilename,int num)
{
  if (1 == output_dir_flag)
    my_snprintf(outfilename,FILENAME_SIZE,"%s/%s%02d.%s",output_directory,prefix,num+offset,op->extension);
  else
    my_snprintf(outfilename,FILENAME_SIZE,"%s%02d.%s",prefix,num+offset,op->extension);
}

static void split_file(wave_info *info)
{
  unsigned char header[CANONICAL_HEADER_SIZE];
  char outfilename[FILENAME_SIZE];
  int current, kidpid_in, kidpid_out;
  wint discard,bytes;
  FILE *output;

  if (NULL == (info->input = info->input_format->open_input(info->filename,&kidpid_in)))
    error("could not reopen file '%s' for input",info->filename);

  discard = info->header_size;
  while (discard > 0) {
    bytes = min(discard,CANONICAL_HEADER_SIZE);
    if (read_n_bytes(info->input,header,bytes) != bytes)
      error("error while discarding %d-byte header",info->header_size);
    discard -= bytes;
  }

  printf("\n");
  printf("Input file '%s' is being split into %d pieces.\n",info->filename,numfiles);
  printf("\n");

  for (current=0;current<numfiles;current++) {
    make_outfilename(outfilename,current);

    if (NULL == (output = op->open_output(outfilename,&kidpid_out)))
      error("could not open file '%s' for output",outfilename);

    files[current]->chunk_size = files[current]->data_size + CANONICAL_HEADER_SIZE - 8;
    files[current]->channels = info->channels;
    files[current]->samples_per_sec = info->samples_per_sec;
    files[current]->avg_bytes_per_sec = info->avg_bytes_per_sec;
    files[current]->block_align = info->block_align;
    files[current]->bits_per_sample = info->bits_per_sample;
    files[current]->wave_format = info->wave_format;
    files[current]->rate = info->rate;
    files[current]->length = files[current]->data_size / (wlong)info->rate;

    make_canonical_header(header,files[current]);

    length_to_str(files[current]);

    printf("Writing %s (%s) ... ",outfilename,files[current]->m_ss);

    fflush(stdout);

    if (write_n_bytes(output,header,CANONICAL_HEADER_SIZE) != CANONICAL_HEADER_SIZE)
      error("error while writing %d-byte header",CANONICAL_HEADER_SIZE);

    if (transfer_n_bytes(info->input,output,files[current]->data_size) != files[current]->data_size)
      error("error while transferring %ld bytes to output file",files[current]->data_size);

    printf("done.\n");

    close_and_wait_output(output,kidpid_out);
  }

  close_and_wait_input(info->input,kidpid_in);
}

static void show_preview(char *filename,wave_info *info)
{
  int i;
  char outfilename[FILENAME_SIZE];

  printf("\n");
  printf("Preview of changes:\n");
  printf("-------------------\n");
  printf("\n");
  printf("File '%s' will be split into %d pieces:\n",filename,numfiles);
  for (i=0;i<numfiles;i++) {
    make_outfilename(outfilename,i);
    printf("\n");

    files[i]->rate = info->rate;
    files[i]->length = files[i]->data_size / (wlong)info->rate;

    length_to_str(files[i]);

    printf("  + %s (%s)\n",outfilename,files[i]->m_ss);
    printf("    - this file will contain %ld bytes of WAVE data (plus %d-byte header)\n",files[i]->data_size,CANONICAL_HEADER_SIZE);
    if (input_is_cd_quality) {
      if (PROB_TOO_SHORT(files[i]))
        printf("    - this file will be too short to be burned\n");
      if (PROB_BAD_BOUND(files[i]))
        printf("    - this file will not be cut on a sector boundary\n");
    }
  }
}

static int is_a_digit(char c)
{
  switch (c) {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      break;
    default:
      return 0;
  }
  return 1;
}

static int is_numeric(unsigned char *buf)
{
  unsigned char *p = buf;
  wlong bytes;

  while ('\0' != *p) {
    if (0 == is_a_digit(*p))
      return 0;
    p++;
  }

  bytes =
#ifdef HAVE_ATOL
      (wlong)atol(buf);
#else
      (wlong)atoi(buf);
#endif

  return bytes;
}

static wlong is_m_ss(unsigned char *buf,wave_info *info)
{
  unsigned char *colon,*p;
  int len,min,sec;
  wlong bytes;

  len = strlen(buf);

  if (len < 4)
    return 0;

  colon = buf + len - 3;

  if (':' != *colon)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';

  min = atoi(buf);
  sec = atoi(colon+1);

  if (sec >= 60)
    error("invalid value for seconds: %d",sec);

  bytes = (wlong)(min * info->rate * 60) +
          (wlong)(sec * info->rate);

  return bytes;
}

static wlong is_m_ss_ff(unsigned char *buf,wave_info *info)
{
  unsigned char *colon,*dot,*p;
  int len,min,sec,frames;
  wlong bytes;

  len = strlen(buf);

  if (len < 7)
    return 0;

  colon = buf + len - 6;
  dot = buf + len - 3;

  if (':' != *colon || '.' != *dot)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon && p != dot) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';
  *dot = '\0';

  if (0 == input_is_cd_quality)
    error("m:ss.ff format can only be used with CD-quality input files");

  min = atoi(buf);
  sec = atoi(colon+1);
  frames = atoi(dot+1);

  if (sec >= 60)
    error("invalid value for seconds: %d",sec);

  if (frames >= 75)
    error("invalid value for frames: %d",frames);

  bytes = (wlong)(min * CD_RATE * 60) +
          (wlong)(sec * CD_RATE) +
          (wlong)(frames * CD_BLOCK_SIZE);

  return bytes;
}

static wlong is_m_ss_nnn(unsigned char *buf,wave_info *info)
{
  unsigned char *colon,*dot,*p;
  int len,min,sec,ms,nearest_byte,nearest_frame;
  wlong bytes;

  len = strlen(buf);

  if (len < 8)
    return 0;

  colon = buf + len - 7;
  dot = buf + len - 4;

  if (':' != *colon || '.' != *dot)
    return 0;

  p = buf;
  while ('\0' != *p) {
    if (p != colon && p != dot) {
      if (0 == is_a_digit(*p))
        return 0;
    }
    p++;
  }

  *colon = '\0';
  *dot = '\0';

  min = atoi(buf);
  sec = atoi(colon+1);
  ms = atoi(dot+1);

  if (sec >= 60)
    error("invalid value for seconds: %d",sec);

  nearest_byte = (int)((((double)ms * (double)info->rate) / 1000.0) + 0.5);

  bytes = (wlong)(min * info->rate * 60) +
          (wlong)(sec * info->rate);

  if (input_is_cd_quality) {
    nearest_frame = (int)((double)ms * .075 + 0.5) * CD_BLOCK_SIZE;
    if (nearest_frame != nearest_byte) {
      warning("rounding %d:%02d.%03d (offset: %lu) to nearest sector boundary (offset: %lu)",
              min,sec,ms,bytes+(wlong)nearest_byte,bytes+(wlong)nearest_frame);
      bytes += (wlong)nearest_frame;
    }
  }

  return bytes;
}

static wlong smrt_parse(unsigned char *data,wave_info *info)
/* currently not so smart, so I gave it the Homer Simpson spelling :) */
{
  wlong bytes;

  while (data[strlen(data)-1] == '\n' ||
         data[strlen(data)-1] == '\r' ||
         data[strlen(data)-1] == '\t' ||
         data[strlen(data)-1] == ' ')
    data[strlen(data)-1] = '\0';

  /* check for all digits */
  if (0 != (bytes = is_numeric(data)))
    return bytes;

  /* check for m:ss */
  if (0 != (bytes = is_m_ss(data,info)))
    return bytes;

  /* check for m:ss.ff */
  if (0 != (bytes = is_m_ss_ff(data,info)))
    return bytes;

  /* check for m:ss.nnn */
  if (0 != (bytes = is_m_ss_nnn(data,info)))
    return bytes;

  /* it was not in any of these formats */

  error("split point '%s' is not in bytes, m:ss, m:ss.ff, or m:ss.nnn format",data);

  return 0;
}

static void read_split_points(wave_info *info)
{
  FILE *fd;
  unsigned char data[BUF_SIZE];
  wlong previous = 0;

  if (1 == split_point_file_flag) {
    if (NULL == (fd = fopen(split_point_file,"rb")))
      error("could not open split point file '%s'",split_point_file);
  }
  else
    fd = stdin;

  fgets(data,BUF_SIZE-1,fd);

  while (0 == feof(fd)) {
    if (MAX_PIECES - 1 == numfiles)
      error("too many split files would be created - maximum is %d",MAX_PIECES);

    if (NULL == (files[numfiles] = new_wave_info(NULL)))
      error("could not allocate memory for split points array");

    files[numfiles]->beginning_byte = smrt_parse(data,info);
    if (files[numfiles]->beginning_byte <= previous)
      error("split point %d comes before previous split point",numfiles+1);

    files[numfiles]->data_size = files[numfiles]->beginning_byte - previous;

    if (input_is_cd_quality) {
      if (files[numfiles]->data_size < CD_MIN_BURNABLE_SIZE) {
        warning("file %d will be too short to be burned",numfiles+1);
        files[numfiles]->problems |= PROBLEM_CD_BUT_TOO_SHORT;
        if (0 != files[numfiles]->data_size % CD_BLOCK_SIZE) {
          files[numfiles]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
          printf("         and will not be cut on a sector boundary\n");
        }
      }
      else if (0 != files[numfiles]->data_size % CD_BLOCK_SIZE) {
        files[numfiles]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
        warning("file %d will not be cut on a sector boundary",numfiles+1);
      }
    }
    else {
      files[numfiles]->problems |= PROBLEM_NOT_CD_QUALITY;
    }

    previous = files[numfiles]->beginning_byte;
    numfiles++;
    fgets(data,BUF_SIZE-1,fd);
  }

  if (NULL == (files[numfiles] = new_wave_info(NULL)))
    error("could not allocate memory for split points array");

  numfiles++;

  if (1 == split_point_file_flag)
    fclose(fd);

  if (1 == numfiles)
    error("no split points given - nothing to do");
}

static void process_file(char *filename)
{
  int i;
  wave_info *info;

  if (NULL == (info = new_wave_info(filename)))
    error("cannot continue due to error(s) shown above");

  input_is_cd_quality = (PROB_NOT_CD(info) ? 0 : 1);

  read_split_points(info);

  if (files[numfiles-2]->beginning_byte > info->data_size)
    error("split points go beyond input file's data size");

  if (files[numfiles-2]->beginning_byte == info->data_size) {
    free(files[numfiles-1]);
    numfiles--;
  }
  else {
    files[numfiles-1]->beginning_byte = info->data_size;
    files[numfiles-1]->data_size = info->data_size - files[numfiles-2]->beginning_byte;

    if (input_is_cd_quality) {
      if (files[numfiles-1]->data_size < CD_MIN_BURNABLE_SIZE) {
        warning("file %d will be too short to be burned",numfiles);
        files[numfiles-1]->problems |= PROBLEM_CD_BUT_TOO_SHORT;
        if (0 != files[numfiles-1]->data_size % CD_BLOCK_SIZE) {
          files[numfiles-1]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
          printf("         and will not be cut on a sector boundary\n");
        }
      }
      else if (0 != files[numfiles-1]->data_size % CD_BLOCK_SIZE) {
        files[numfiles-1]->problems |= PROBLEM_CD_BUT_BAD_BOUND;
        warning("file %d will not be cut on a sector boundary",numfiles);
      }
    }
    else {
      files[numfiles-1]->problems |= PROBLEM_NOT_CD_QUALITY;
    }
  }

  if (1 == preview)
    show_preview(info->filename,info);
  else
    split_file(info);

  for (i=0;i<numfiles;i++)
    free(files[i]);

  free(info);
}

static void process(int argc,char **argv,int start)
{
  char filename[FILENAME_SIZE];

  if (argc < start + 1) {
    /* no filenames were given, so we're reading one file from standard input. */
    fgets(filename,FILENAME_SIZE-1,stdin);
    if (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file(filename);
    }
  }
  else
    process_file(argv[start]);
}

static void split_main(int argc,char **argv,int start)
{
  int first_arg;

  parse(argc,argv,start,&first_arg);

  process(argc,argv,first_arg);
}
