/*  mode_strip.c - strip mode module
 *  Copyright (C) 2000-2003  Jason Jordan <shnutils@freeshell.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: mode_strip.c,v 1.18 2003/02/17 04:10:42 jason Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shntool.h"
#include "convert.h"
#include "fileio.h"
#include "wave.h"
#include "misc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define POSTFIX_STRIP "-stripped"

static int output_dir_flag = 0;
static int strip_header = 1;
static int strip_chunks = 1;
static int preview = 0;

static char *output_directory = ".";

static format_module *op = NULL;

static void strip_main(int,char **,int);

mode_module mode_strip = {
  "strip",
  "shnstrip",
  "strips extra RIFF chunks and/or writes canonical headers",
  strip_main
};

static void show_usage()
{
  int i,j=0;

  printf("Usage: %s [OPTIONS] [files]\n",fullprogname);
  printf("\n");
  printf("Options:\n");
  printf("\n");
  printf("  -o fmt  specifies the output file format, where fmt is one of the following:\n");
  printf("\n");
  for (i=0;formats[i];i++)
    if (formats[i]->open_output)
      printf("             %" MAX_MODULE_NAME_LENGTH "s  (%s) %s\n",formats[i]->name,formats[i]->description,(0 == (j++))?"[DEFAULT]":"");
  printf("\n");
  printf("  -d dir  specifies alternate output directory (default is same as input file)\n");
  printf("\n");
  printf("  -nh     don't rewrite WAVE header in canonical format (default is yes)\n");
  printf("  -nc     don't strip unnecessary RIFF chunks (default is yes)\n");
  printf("\n");
  printf("  -p      preview what changes would be made without actually making them\n");
  printf("\n");
  printf("  --      indicates that everything following it is a file name\n");
  printf("\n");
  printf("  -v      shows version information\n");
  printf("  -h      shows this help screen\n");
  printf("\n");
  printf("  If no file names are given, then file names are read from standard input.\n");
  printf("\n");
  printf("  ------------\n");
  printf("  Sample uses:\n");
  printf("  ------------\n");
  printf("\n");
  printf("  %% %s -o shn -d /mnt/shn/stripped *.shn *.wav\n",fullprogname);
  printf("\n");
  printf("  %% find /your/audio/dir -type f | %s\n",fullprogname);
  printf("\n");

  exit(0);
}

static void parse(int argc,char **argv,int begin,int *first_arg)
{
  int start,i,found_output = -1;

  start = begin;

  for (i=0;formats[i];i++)
    if (formats[i]->open_output) {
      found_output = i;
      break;
    }

  if (-1 == found_output)
    error("no output formats found, cannot continue");

  while (start < argc && argv[start][0] == '-') {
    if (argc > start && 0 == strcmp(argv[start],"--")) {
      start++;
      break;
    }
    else if (argc > start && 0 == strcmp(argv[start],"-v"))
      internal_version();
    else if (argc > start && 0 == strcmp(argv[start],"-h"))
      show_usage();
    else if (argc > start && 0 == strcmp(argv[start],"-nh"))
      strip_header = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-nc"))
      strip_chunks = 0;
    else if (argc > start && 0 == strcmp(argv[start],"-p"))
      preview = 1;
    else if (argc > start && 0 == strcmp(argv[start],"-d")) {
      if (1 == output_dir_flag)
        help("too many output directories specified");
      start++;
      output_dir_flag = 1;
      if (argc <= start)
        help("missing output directory");
      output_directory = argv[start];
    }
    else if (argc > start && 0 == strcmp(argv[start],"-o")) {
      if (NULL != op)
        help("too many output file formats specified");
      op = output_format_init(argc,argv,&start);
    }
    else {
      help("unknown argument: %s",argv[start]);
    }
    start++;
  }

  if (0 == strip_header && 0 == strip_chunks)
    help("nothing to strip if both -nh and -nc are specified");

  if (NULL == op) {
    op = formats[found_output];
    warning("no output format specified - defaulting to %s",op->name);
  }

  *first_arg = start;
}

static void make_outfilename(char *infilename,char *outfilename)
{
  char *basename;

  if ((basename = strrchr(infilename,'/')))
    basename++;
  else
    basename = infilename;

  if (1 == output_dir_flag)
    my_snprintf(outfilename,FILENAME_SIZE,"%s/%s",output_directory,basename);
  else
    my_snprintf(outfilename,FILENAME_SIZE,"%s",infilename);

  if (filename_contains_a_dot(outfilename))
    *(strrchr(outfilename,'.')) = '\0';

  strcat(outfilename,POSTFIX_STRIP);
  strcat(outfilename,".");
  strcat(outfilename,op->extension);
}

static void strip_and_canonicize(wave_info *info)
{
  wint new_header_size;
  wlong new_chunk_size,possible_extra_stuff;
  unsigned char *header;
  char outfilename[FILENAME_SIZE];
  FILE *output;
  int kidpid_in, kidpid_out;

  if (PROB_HDR_INCONSISTENT(info)) {
    warning("skipping file '%s' because its header is inconsistent",info->filename);
    return;
  }

  if (PROB_TRUNCATED(info)) {
    warning("skipping file '%s' because it seems to be truncated",info->filename);
    return;
  }

  if ((1 == strip_header && 1 == strip_chunks) && ((0 == PROB_EXTRA_CHUNKS(info)) && (0 == PROB_HDR_NOT_CANONICAL(info)))) {
    warning("skipping file '%s' because it already has a canonical header and no extra RIFF chunks",info->filename);
    return;
  }

  if ((1 == strip_chunks && 0 == strip_header) && (0 == PROB_EXTRA_CHUNKS(info))) {
    warning("skipping file '%s' because it already has no extra RIFF chunks",info->filename);
    return;
  }

  if ((1 == strip_header && 0 == strip_chunks) && (0 == PROB_HDR_NOT_CANONICAL(info))) {
    warning("skipping file '%s' because it already has a canonical header",info->filename);
    return;
  }

  new_header_size = info->header_size;
  new_chunk_size = info->chunk_size - (info->header_size - CANONICAL_HEADER_SIZE);
  possible_extra_stuff = info->total_size - (info->data_size + info->header_size);

  make_outfilename(info->filename,outfilename);

  if (1 == preview) {
    printf("%s --> %s\n",info->filename,outfilename);
    if (1 == strip_header && PROB_HDR_NOT_CANONICAL(info))
      printf("  - will rewrite %d-byte WAVE header to the canonical %d-byte header\n",info->header_size,CANONICAL_HEADER_SIZE);
    if (1 == strip_chunks && PROB_EXTRA_CHUNKS(info))
      printf("  - will strip %lu byte%s worth of extra RIFF chunk(s) from the end of this file\n",possible_extra_stuff,(possible_extra_stuff > 1)?"s":"");
    printf("\n");
    return;
  }

  if (1 == strip_header)
    new_header_size = CANONICAL_HEADER_SIZE;

  if (1 == strip_chunks) {
    new_chunk_size = (long)new_header_size + info->data_size - 8;
    possible_extra_stuff = 0;
  }
  info->new_data_size = info->data_size + possible_extra_stuff;

  if (NULL == (info->input = info->input_format->open_input(info->filename,&kidpid_in))) {
    warning("could not reopen file '%s' for input - skipping this file",info->filename);
    return;
  }

  if (NULL == (output = op->open_output(outfilename,&kidpid_out))) {
    warning("could not open file '%s' for output - skipping this file",outfilename);
    return;
  }

  if (NULL == (header = malloc(info->header_size * sizeof(unsigned char)))) {
    warning("could not allocate %d bytes for WAVE header - skipping this file",info->header_size);
    return;
  }

  if (read_n_bytes(info->input,header,info->header_size) != info->header_size) {
    warning("error while reading %d bytes of data - skipping this file",info->header_size);
    return;
  }

  /* already read in header, now check to see if we need to overwrite that with a canonical one */
  if (1 == strip_header)
    make_canonical_header(header,info);

  ulong_to_uchar_le(header+4,new_chunk_size);

  if (write_n_bytes(output,header,new_header_size) != new_header_size) {
    warning("error while writing %d bytes of data - skipping this file",new_header_size);
    return;
  }

  printf("%s --> %s ... ",info->filename,outfilename);
  fflush(stdout);

  if (transfer_n_bytes(info->input,output,info->new_data_size) != info->new_data_size) {
    warning("error while writing %lu bytes of data - skipping this file",info->new_data_size);
    return;
  }

  printf("done.\n");

  free(header);

  close_and_wait_input(info->input,kidpid_in);
  close_and_wait_output(output,kidpid_out);
}

static void process_file(char *filename)
{
  wave_info *info;

  if (NULL == (info = new_wave_info(filename)))
    return;

  strip_and_canonicize(info);

  free(info);
}

static void process(int argc,char **argv,int start)
{
  int i;
  char filename[FILENAME_SIZE];

  if (1 == preview) {
    printf("\n");
    printf("Preview of changes:\n");
    printf("-------------------\n");
    printf("\n");
  }
  if (argc < start + 1) {
    /* no filenames were given, so we're reading files from standard input.  */
    fgets(filename,FILENAME_SIZE-1,stdin);
    while (0 == feof(stdin)) {
      filename[strlen(filename)-1] = '\0';
      process_file(filename);
      fgets(filename,FILENAME_SIZE-1,stdin);
    }
  }
  else
    for (i=start;i<argc;i++)
      process_file(argv[i]);
}

static void strip_main(int argc,char **argv,int start)
{
  int first_arg;

  parse(argc,argv,start,&first_arg);

  process(argc,argv,first_arg);
}
